@@ -61,7 +61,7 @@
 #define MINVERSION 0 /* API compatible, ABI may change, functional
 		      * enhancements only, consumer can be left unchanged if
 		      * enhancements are not considered */
-#define PATCHLEVEL 3 /* API / ABI compatible, no functional changes, no
+#define PATCHLEVEL 4 /* API / ABI compatible, no functional changes, no
 		      * enhancements, bug fixes only */
 
 static int Verbosity = 0;
@@ -96,9 +96,11 @@ static char *Pidfile = NULL;
 
 static int Entropy_avail_fd = 0;
 
-#define RNDBYTES 256
+#define ENTROPYBYTES 32
+#define OVERSAMPLINGFACTOR 2
 #define ENTROPYTRHESH 1024
 #define ENTROPYAVAIL "/proc/sys/kernel/random/entropy_avail"
+#define LRNG_FILE "/proc/sys/kernel/random/lrng_type"
 
 static void install_alarm(void);
 static void dealloc(void);
@@ -204,12 +206,14 @@ static inline void memset_secure(void *s, int c, size_t n)
  * entropy handler functions
  *******************************************************************/
 
-static size_t write_random(struct kernel_rng *rng, char *buf, size_t len)
+static size_t write_random(struct kernel_rng *rng, char *buf, size_t len,
+			   size_t entropy_bytes)
 {
 	size_t written = 0;
 	int ret;
 
-	rng->rpi->entropy_count = (len * 8); /* value is in bits */
+	 /* value is in bits */
+	rng->rpi->entropy_count = (entropy_bytes * 8);
 	rng->rpi->buf_size = len;
 	memcpy(rng->rpi->buf, buf, len);
 	memset(buf, 0, len);
@@ -218,7 +222,8 @@ static size_t write_random(struct kernel_rng *rng, char *buf, size_t len)
 	if (0 > ret)
 		dolog(LOG_WARN, "Error injecting entropy: %s", strerror(errno));
 	else {
-		dolog(LOG_DEBUG, "Injected %d bytes of entropy", len);
+		dolog(LOG_DEBUG, "Injected %u bytes with an entropy contnt of %u bytes of entropy",
+		      len, entropy_bytes);
 		written = len;
 	}
 
@@ -231,20 +236,20 @@ static size_t write_random(struct kernel_rng *rng, char *buf, size_t len)
 
 static size_t gather_entropy(struct kernel_rng *rng)
 {
-	char buf[RNDBYTES];
+	char buf[(ENTROPYBYTES * OVERSAMPLINGFACTOR)];
 	size_t ret = 0;
 
-	if (0 > jent_read_entropy(rng->ec, buf, RNDBYTES)) {
+	if (0 > jent_read_entropy(rng->ec, buf, sizeof(buf))) {
 		dolog(LOG_WARN, "Cannot read entropy");
 		return 0;
 	}
-	ret = write_random(rng, buf, RNDBYTES);
-	if (RNDBYTES != ret)
+	ret = write_random(rng, buf, sizeof(buf), ENTROPYBYTES);
+	if (sizeof(buf) != ret)
 		dolog(LOG_WARN, "Injected %lu bytes into %s, expected %d",
-			ret, rng->dev, RNDBYTES);
-	memset_secure(buf, 0, RNDBYTES);
+			ret, rng->dev, sizeof(buf));
+	memset_secure(buf, 0, sizeof(buf));
 
-	return RNDBYTES;
+	return sizeof(buf);
 }
 
 static int read_entropy_avail(int fd)
@@ -306,6 +311,17 @@ out:
 	return;
 }
 
+/* Is the LRNG present instead of the legacy /dev/random? */
+static int lrng_present(void)
+{
+	struct stat buf;
+	int ret = stat(LRNG_FILE, &buf);
+
+	if (ret == -1 && errno == ENOENT)
+		return 0;
+	return 1;
+}
+
 /* terminate the daemon cleanly */
 static void sig_term(int sig)
 {
@@ -343,6 +359,8 @@ static void select_fd(void)
 
 static void install_alarm(void)
 {
+	if (lrng_present())
+		return;
 	dolog(LOG_DEBUG, "Install alarm signal handler");
 	signal(SIGALRM, sig_entropy_avail);
 	alarm(5);
@@ -368,7 +386,7 @@ static void alloc_rng(struct kernel_rng *rng)
 		dolog(LOG_ERR, "Allocation of entropy collector failed");
 
 	rng->rpi = malloc((sizeof(struct rand_pool_info) +
-			  (RNDBYTES * sizeof(char))));
+			  (ENTROPYBYTES * OVERSAMPLINGFACTOR * sizeof(char))));
 	if (!rng->rpi)
 		dolog(LOG_ERR, "Cannot allocate memory for random bytes");
 
@@ -404,7 +422,8 @@ static void dealloc_rng(struct kernel_rng *rng)
 	}
 	if (NULL != rng->rpi) {
 		memset(rng->rpi, 0,(sizeof(struct rand_pool_info) +
-				    (RNDBYTES * sizeof(char))));
+				    (ENTROPYBYTES * OVERSAMPLINGFACTOR *
+				     sizeof(char))));
 		free(rng->rpi);
 		rng->rpi = NULL;
 	}
